// Deliberately leaving just to JIT, because AOT will have complications of its own, like:
//  - variadic arguments
//  - passing X/Y coordinates along with pictures
//  - the output of intermediate stages (debug_to_file produces empty images right now)

#include "Halide.h"
#include "myimg.h"

#define	COLOURS	4	/* RGBA */
#define	BYTES	1	/* 8-bit */

template <typename T>
Halide::Image<T>
blend(Halide::Image<T> &top, Halide::Image<T> &left, Halide::Image<T> &right,
    Halide::Image<T> &bottom, box &overlap, box &un)
{
	assert(&top != &bottom && &left != &right);

	Halide::Var x("x"), y("y"), c("c");
	Halide::Func concat("concat");

	assert(left.stride(0) == 1);  // Halide secretly needs that.
	assert(left.stride(1) == COLOURS);  // The X dimension happens every RGBA.
	printf("copying non-overlap sections...\n");

	concat(c, x, y) = Halide::cast<T>(0);

	/* How far down the Y axis are left and right images shifted? */
	int yoffL = (&top == &left) ? 0 : right.extent(2) - overlap.h;
	int yoffR = (&top == &right) ? 0 : left.extent(2) - overlap.h;

	/* How far in the X axis are top + bottom non-overlap regions placed? */
	int xoffT = (&top == &left) ? top.extent(1) - overlap.w : 0;
	int xoffB = (&top == &right) ? bottom.extent(1) - overlap.w : 0;

	/* Copy non-overlap pixels on the X dimension. */
	int xoff = left.extent(1) - overlap.w;
	Halide::RDom lrd(0, xoff, 0, left.extent(2));
	Halide::RDom rrd(overlap.w, right.extent(1) - overlap.w, 0, right.extent(2));
	printf("X dimension non-overlap L: %d,\tR: %d\n",
	    left.extent(1) - overlap.w, right.extent(1) - overlap.w);

	concat(c, lrd.x,        lrd.y + yoffL) = left(c, lrd.x, lrd.y);
	concat(c, rrd.x + xoff, rrd.y + yoffR) = right(c, rrd.x, rrd.y);

	/* Copy non-overlap pixels on the Y dimension. */
	Halide::RDom trd(xoffT, overlap.w, 0, top.extent(2) - overlap.h);
	Halide::RDom brd(xoffB, overlap.w,
	    overlap.h, bottom.extent(2) - overlap.h);
	printf("Y dimension non-overlap T: %d,\tB: %d\n",
	    top.extent(2) - overlap.h, bottom.extent(2) - overlap.h);

	int xoffTdst = (&top == &left) ? 0 : xoff;
	int xoffBdst = (&top == &left) ? xoff : 0;
	int yoffB = (&bottom == &left) ? yoffL : yoffR;

	concat(c, trd.x + xoffTdst, trd.y) = top(c, trd.x, trd.y);
	concat(c, brd.x + xoffBdst, brd.y + yoffB) = bottom(c, brd.x, brd.y);

	/* Overlap cross-section computation goes here. */
	printf("blending the cross-section...\n");
	Halide::RDom xs({{0, overlap.w}, {0, overlap.h}});

	/* Fill the xsection with whatever is not transparent.  No pyramids. */
	// XXX explain and comment unintuitive +offL in R and vice versa
	concat(c, xs.x + xoff, xs.y + yoffB) =
	    select(right(3, xs.x, xs.y + yoffL) == 255,
	    right(c, xs.x, xs.y + yoffL),
	    left(c, xs.x + xoff, xs.y + yoffR));

	concat.output_buffer().set_stride(1, left.stride(1))
	    .set_bounds(0, 0, COLOURS);
	printf("realizing output...\n");

	Halide::Image<T> output;
	output = concat.realize(COLOURS, un.w, un.h);
	return output;
}

template <typename T>
Halide::Image<T>
blend_myimg(myimg &top, myimg &left, myimg &right, box &un)
{
	struct buffer_t & lbuf = left.prepare_buffer();
	struct buffer_t & rbuf = right.prepare_buffer();

	Halide::Image<T> l(&lbuf, "left"), r(&rbuf, "right");
	Halide::Image<T> & t = (&top == &left) ? l : r;
	Halide::Image<T> & b = (&top == &left) ? r : l;

	auto overlap = left.overlap(right);
	un = left.union_with(right);
	printf("%d %d %d %d\n", overlap.x, overlap.y, overlap.w, overlap.h);
	printf("%d %d %d %d\n", un.x, un.y, un.w, un.h);

	return blend(t, l, r, b, overlap, un);
}

void
stitch(const std::string dst, const std::string src1, const std::string src2)
{
	TIFFimg aimg(src1), bimg(src2);

	/* Figure out the geometry */
	printf("%s is on the left\n", aimg.left_from(bimg) ? src1.c_str() : src2.c_str());
	printf("%s is on the top\n", aimg.above(bimg) ? src1.c_str() : src2.c_str());

	Halide::Image<uint8_t> output;
	box un;
	if (aimg.left_from(bimg)) {
		if (aimg.above(bimg)) {
			output = blend_myimg<uint8_t>(aimg, aimg, bimg, un);
		} else {
			output = blend_myimg<uint8_t>(bimg, aimg, bimg, un);
		}
	} else {
		if (aimg.above(bimg)) {
			output = blend_myimg<uint8_t>(aimg, bimg, aimg, un);
		} else {
			output = blend_myimg<uint8_t>(bimg, bimg, aimg, un);
		}
	}
	TIFFimg out(dst, output.raw_buffer(), un.x, un.y);
}
